@@ -1,8 +1,9 @@
 #!/usr/bin/env python3
 # -*- coding: utf-8 -*-
 
-from .core import * 
-from .core import _open_source
 import pkg_resources
+from .core import *
 
 __version__ = pkg_resources.require("dessia_common")[0].version
+
+FLOAT_TOLERANCE = 1e-9
