@@ -6,20 +6,20 @@ Created on Wed Feb 19 15:56:12 2020
 @author: jezequel
 """
 
+from io import StringIO
 import math
-from typing import List, Dict
+from typing import List, Dict, Any
 import numpy as np
 import pandas as pd
-from dessia_common import DessiaObject, Parameter, is_bounded, Filter
+from dessia_common import DessiaObject, Parameter, is_bounded, DessiaFilter
 from scipy.optimize import minimize
 from pyDOE import lhs
 
 
 class ParetoSettings(DessiaObject):
-    """ 
-    :param coeff_names: A dictionary containing the name of a variable as key and \
-    a boolean for minimization as value.
-    :type coeff_names: Dict[str, bool]
+    """
+    :param minimized_attributes: A dictionary containing the name of a variable as key
+                                 and a boolean for minimization as value.
     :param enabled: List of strings representing ordered variables names.
     :type enabled: bool
     :param name: The name of the block.
@@ -28,8 +28,7 @@ class ParetoSettings(DessiaObject):
     _generic_eq = True
     _ordered_attributes = ['name', 'enabled', 'minimized_attributes']
 
-    def __init__(self, minimized_attributes: Dict[str, bool],
-                 enabled: bool = True, name: str = ''):
+    def __init__(self, minimized_attributes: Dict[str, bool], enabled: bool = True, name: str = ''):
         self.enabled = enabled
         self.minimized_attributes = minimized_attributes
 
@@ -40,8 +39,7 @@ class ObjectiveSettings(DessiaObject):
     _generic_eq = True
     _ordered_attributes = ['name', 'enabled', 'n_near_values']
 
-    def __init__(self, n_near_values: int = 4,
-                 enabled: bool = True, name: str = ''):
+    def __init__(self, n_near_values: int = 4, enabled: bool = True, name: str = ''):
         self.n_near_values = n_near_values
         self.enabled = enabled
 
@@ -52,11 +50,6 @@ class Objective(DessiaObject):
     """
     Defines an objective function
 
-    :param coeff_names: List of strings representing ordered variables names.
-    :type coeff_names: [str]
-    :param coeff_values: List of floats representing coefficients.
-    :type coeff_values: [float]
-
     Order is kept. It means that coefficients are applied to the variables
     in the same order as they are defined.
     """
@@ -65,14 +58,12 @@ class Objective(DessiaObject):
     _ordered_attributes = ['name', 'settings', 'coefficients']
     _non_serializable_attributes = ['coeff_names']
 
-    def __init__(self, coefficients: Dict[str, float],
-                 directions: Dict[str, bool],
+    def __init__(self, coefficients: Dict[str, float], directions: Dict[str, bool],
                  settings: ObjectiveSettings, name: str = ''):
         for variable in coefficients:
             if variable not in directions:
-                msg = "Coefficient for variable {}".format(variable)
-                msg += " was found but no direction is specified."
-                msg += " Add {} to directions dict.".format(variable)
+                msg = f"Coefficient for variable {variable} was found but no direction is specified." \
+                      f"Add {variable} to directions dict."
                 raise KeyError(msg)
         self.coefficients = coefficients
         self.directions = directions
@@ -85,7 +76,7 @@ class Objective(DessiaObject):
         rating = 0
         for variable, value in values.items():
             coefficient = self.coefficients[variable]
-            rating += coefficient*value
+            rating += coefficient * value
         return rating
 
     @classmethod
@@ -98,9 +89,9 @@ class Objective(DessiaObject):
         for i, angle in enumerate(angles):
             matrix_i = np.identity(n)
             matrix_i[i, i] = math.cos(angle)
-            matrix_i[i+1, i] = -math.sin(angle)
-            matrix_i[i, i+1] = math.sin(angle)
-            matrix_i[i+1, i+1] = math.cos(angle)
+            matrix_i[i + 1, i] = -math.sin(angle)
+            matrix_i[i, i + 1] = math.sin(angle)
+            matrix_i[i + 1, i + 1] = math.cos(angle)
             matrix = np.dot(matrix, matrix_i)
         x = np.zeros(n)
         x[0] = 1
@@ -109,10 +100,8 @@ class Objective(DessiaObject):
         return unsigned
 
     @classmethod
-    def from_angles(cls, angles: List[float], variables: List[str],
-                    directions: Dict[str, bool],
-                    settings: ObjectiveSettings = None,
-                    name="Generated from angles") -> 'Objective':
+    def from_angles(cls, angles: List[float], variables: List[str], directions: Dict[str, bool],
+                    settings: ObjectiveSettings = None, name="Generated from angles") -> 'Objective':
         if not isinstance(angles, list) and not isinstance(angles, np.ndarray):
             angles = [angles]
         generated_coefficients = cls.coefficients_from_angles(angles=angles)
@@ -128,8 +117,7 @@ class Objective(DessiaObject):
         if settings is None:
             settings = ObjectiveSettings()
 
-        return Objective(coefficients=coefficients, directions=directions,
-                         settings=settings, name=name)
+        return Objective(coefficients=coefficients, directions=directions, settings=settings, name=name)
 
 
 class Catalog(DessiaObject):
@@ -140,8 +128,6 @@ class Catalog(DessiaObject):
     :param objectives: List of objectives to apply to
                        catalog vectored objects
     :type objectives: [Objective]
-    # :param objects: List of vectored objects.
-    # :type objects: [VectoredObject]
     :param choice_variables: List of string. List of variable names
                              that represent choice arguments
     :type choice_variables: [str]
@@ -152,17 +138,12 @@ class Catalog(DessiaObject):
     _generic_eq = True
     _standalone_in_db = True
     _ordered_attributes = ['name', 'pareto_settings', 'objectives']
-    _non_editable_attributes = ['array', 'variables', 'choice_variables',
-                                'generated_best_objectives']
-    _export_formats = ['csv']
+    _non_editable_attributes = ['array', 'variables', 'choice_variables', 'generated_best_objectives']
     _allowed_methods = ['find_best_objective']
     _whitelist_attributes = ['variables']
 
-    def __init__(self, array: List[List[float]], variables: List[str],
-                 pareto_settings: ParetoSettings,
-                 objectives: List[Objective] = None,
-                 choice_variables: List[str] = None,
-                 name: str = ''):
+    def __init__(self, array: List[List[float]], variables: List[str], pareto_settings: ParetoSettings = None,
+                 objectives: List[Objective] = None, choice_variables: List[str] = None, name: str = ''):
         DessiaObject.__init__(self, name=name)
 
         self.array = array
@@ -182,17 +163,14 @@ class Catalog(DessiaObject):
         self.generated_best_objectives = 0
 
     @classmethod
-    def concatenate(cls, catalogs: List['Catalog'],
-                    pareto_settings: ParetoSettings = None,
-                    objectives: List[Objective] = None,
-                    choice_variables: List[str] = None,
+    def concatenate(cls, catalogs: List['Catalog'], pareto_settings: ParetoSettings = None,
+                    objectives: List[Objective] = None, choice_variables: List[str] = None,
                     name: str = '') -> 'Catalog':
         varsets = [set(c.variables) for c in catalogs]
         var_intersection = list(set.intersection(*varsets))
 
         if choice_variables is not None:
-            choice_variables = [var for var in choice_variables
-                                if var in var_intersection]
+            choice_variables = [var for var in choice_variables if var in var_intersection]
         else:
             choice_variables = var_intersection
 
@@ -200,130 +178,54 @@ class Catalog(DessiaObject):
         for cat in catalogs:
             indices = [cat.get_variable_index(v) for v in var_intersection]
             for line in cat.array:
-                line_intersection = [v for j, v in enumerate(line)
-                                     if j in indices]
+                line_intersection = [v for j, v in enumerate(line) if j in indices]
                 array_intersection.append(line_intersection)
 
         if pareto_settings is None:
             min_attrs = {var: True for var in choice_variables}
-            pareto_settings = ParetoSettings(minimized_attributes=min_attrs,
-                                             enabled=False)
+            pareto_settings = ParetoSettings(minimized_attributes=min_attrs, enabled=False)
 
-        catalog = cls(array=array_intersection, variables=var_intersection,
-                      pareto_settings=pareto_settings, objectives=objectives,
-                      choice_variables=choice_variables, name=name)
+        catalog = cls(array=array_intersection, variables=var_intersection, pareto_settings=pareto_settings,
+                      objectives=objectives, choice_variables=choice_variables, name=name)
         return catalog
 
-    # def __getitem__(self, item):
-    #     print('Getitem', item)
-    #     return self.array[item]
-    #
-    # def __getattr__(self, item):
-    #     if isinstance(item, (list, tuple)):
-    #         return enhanced_deep_attr(self, item)w
-    #     elif isinstance(item, str) and '.' in item:
-    #         sequence = item.split('.')
-    #         healed_sequence = []
-    #         for i, attr in enumerate(sequence):
-    #             try:
-    #                 healed_sequence.append(int(attr))
-    #             except ValueError:
-    #                 if attr in self.variables:
-    #                     healed_sequence.append(self.get_variable_index(attr))
-    #                 else:
-    #                     healed_sequence.append(attr)
-    #         print(healed_sequence)
-    #         return enhanced_deep_attr(self, healed_sequence)
-    #     return self.__getattribute__(item)
+    def generate_multiplot(self, values: Dict[str, Any] = None):
+        # TOCHECK Avoid circular imports
+        import plot_data
 
-    def _display_angular(self):
-        """
-        Configures catalog display on frontend
+        if values is None:
+            values = []
+            for line in self.array:
+                value = {}
+                for variable in self.variables:
+                    value[variable] = self.get_value_by_name(line, variable)
+                values.append(value)
 
-        :return: List of displays dictionnaries
-        """
-        filters = [variable for j, variable in enumerate(self.variables)
-                   if not isinstance(self.array[0][j], str)
-                   and variable in self.choice_variables]
+        first_vars = self.variables[:2]
+        values2d = [{key: val[key]} for key in first_vars for val in values]
+        rgbs = [[192, 11, 11], [14, 192, 11], [11, 11, 192]]
 
-        # Pareto
-        if self.pareto_settings.enabled:
-            costs = self.build_costs(self.pareto_settings)
-            pareto_indices = pareto_frontier(costs=costs)
-        else:
-            pareto_indices = []
+        tooltip = plot_data.Tooltip(attributes=self.variables, name='Tooltip')
 
-        all_near_indices = {}
-        objective_ratings = {}
-        for iobjective, objective in enumerate(self.objectives):
-            if objective.settings.enabled:
-                ratings = self.handle_objective(objective)
-                if objective.name and objective.name not in objective_ratings:
-                    name = objective.name
-                else:
-                    name = 'objective_'+str(iobjective)
-                objective_ratings[name] = ratings
-                filters.append(name)
-                nval = objective.settings.n_near_values
-                near_indices = list(np.argpartition(ratings, nval)[:nval])
-                all_near_indices[iobjective] = near_indices
+        scatterplot = plot_data.Scatter(axis=plot_data.Axis(), tooltip=tooltip, x_variable=first_vars[0],
+                                        y_variable=first_vars[1], elements=values2d, name='Scatter Plot')
 
-        datasets = []
+        parallelplot = plot_data.ParallelPlot(disposition='horizontal', axes=self.variables, rgbs=rgbs, elements=values)
+        objects = [scatterplot, parallelplot]
+        sizes = [plot_data.Window(width=560, height=300),
+                 plot_data.Window(width=560, height=300)]
+        coords = [(0, 0), (0, 300)]
+        multiplot = plot_data.MultiplePlots(plots=objects, elements=values, sizes=sizes,
+                                            coords=coords, name='Results plot')
+        return multiplot
 
-        values = []
-        for i, line in enumerate(self.array):
-            value = {}
-            for variable in self.variables:
-                value[variable] = self.get_value_by_name(line, variable)
-            for objective_name, ratings in objective_ratings.items():
-                value[objective_name] = ratings[i]
-            values.append(value)
-
-        # Pareto
-        if self.pareto_settings.enabled:
-            pareto_points = [i for i in range(len(values))
-                             if pareto_indices[i]]
-            datasets.append({'label': 'Pareto frontier',
-                             'color': '#ffcc00',
-                             'values': pareto_points})
-
-        # Objectives
-        for iobjective, near_indices in all_near_indices.items():
-            objective = self.objectives[iobjective]
-            if objective.name:
-                label = objective.name
-            else:
-                label = 'Near Values ' + str(iobjective)
-            near_points = [i for i in range(len(values)) if i in near_indices]
-            near_dataset = {'label': label,
-                            'color': None,
-                            'values': near_points}
-            datasets.append(near_dataset)
-
-        # Dominated points
-        dominated_points = [i for i in range(len(values))
-                            if (self.pareto_settings.enabled
-                                and not pareto_indices[i]
-                                or not self.pareto_settings.enabled)]
-        datasets.append({'label': 'Dominated points',
-                         'color': "#99b4d6",
-                         'values': dominated_points})
-
-        # Displays
-        displays = [{'angular_component': 'results',
-                     'filters': filters,
-                     'datasets': datasets,
-                     'values': values,
-                     'references_attribute': 'array'}]
-        return displays
-
-    def filter_(self, filters: List[Filter]):
+    def filter_(self, filters: List[DessiaFilter]):
         def apply_filters(line):
             bounded = True
             i = 0
             while bounded and i < len(filters):
                 filter_ = filters[i]
-                variable = filter_['attribute']
+                variable = filter_.attribute
                 value = line[self.get_variable_index(variable)]
                 bounded = is_bounded(filter_, value)
                 i += 1
@@ -332,6 +234,22 @@ class Catalog(DessiaObject):
         filtered_array = list(filter(apply_filters, self.array))
         return filtered_array
 
+    @classmethod
+    def from_csv(cls, file: StringIO, end: int = None, remove_duplicates: bool = False):
+        """
+        Generates MBSEs from given .csv file.
+        """
+        array = np.genfromtxt(file, dtype=None, delimiter=',', names=True, encoding=None)
+        variables = [v for v in array.dtype.fields.keys()]
+        lines = []
+        for i, line in enumerate(array):
+            if end is not None and i >= end:
+                break
+            if not remove_duplicates or (remove_duplicates
+                                         and line.tolist() not in lines):
+                lines.append(line.tolist())
+        return cls(lines, variables)
+
     def export_csv(self, attribute_name: str, indices: List[int], file: str):
         """
         Exports a reduced list of objects to .csv file
@@ -371,16 +289,21 @@ class Catalog(DessiaObject):
     def get_variable_index(self, name: str) -> int:
         return self.variables.index(name)
 
-    def get_values(self, variable: str) -> List[float]:
+    def get_values(self, variable: str) -> List:
         values = [self.get_value_by_name(line, variable)
                   for line in self.array]
         return values
 
-    def get_value_by_name(self, line: List[float], name: str) -> float:
+    def get_value_by_name(self, line: List, name: str) -> float:
         j = self.get_variable_index(name)
         value = line[j]
         return value
 
+    @property
+    def pareto_is_enabled(self) -> bool:
+        pareto = self.pareto_settings
+        return pareto is not None and pareto.enabled
+
     def build_costs(self, pareto_settings: ParetoSettings):
         """
         Build list of costs that are used to compute Pareto frontier.
@@ -426,9 +349,9 @@ class Catalog(DessiaObject):
                                          enabled=True)
         objectives = []
         while len(array) <= end:
-            line = [(bounds[v][1]-bounds[v][0])*np.random.rand() + bounds[v][0]
+            line = [(bounds[v][1] - bounds[v][0]) * np.random.rand() + bounds[v][0]
                     for v in variables]
-            if line[0]*line[1] >= threshold:
+            if line[0] * line[1] >= threshold:
                 array.append(line)
         return cls(array=array, variables=variables,
                    pareto_settings=pareto_settings, objectives=objectives,
@@ -491,13 +414,98 @@ class Catalog(DessiaObject):
         if not success:
             raise ValueError("No solutions found")
 
+    def plot_data(self):
+        from plot_data.core import Tooltip, TextStyle, SurfaceStyle,\
+            Scatter, ParallelPlot, PointFamily, MultiplePlots
+        from plot_data.colors import GREY, LIGHTGREY, LIGHTGREEN, LIGHTBLUE
+
+        name_column_0 = self.variables[0]
+        list_name = self.get_values(name_column_0)
+
+        list_settings = [name
+                         for name in self.pareto_settings.minimized_attributes]
+        list_value = [self.get_values(cv) for cv in self.choice_variables]
+        if self.pareto_is_enabled:
+            cost = self.build_costs(self.pareto_settings)
+            p_frontier = pareto_frontier(cost)
+            elements = [[], []]  # point_non_pareto, point_pareto
+            for i, name in enumerate(list_name):
+                dict_element = {name_column_0: name}
+                for k, setting in enumerate(self.choice_variables):
+                    dict_element[setting] = list_value[k][i]
+
+                if p_frontier[i]:
+                    elements[1].append(dict_element)
+                else:
+                    elements[0].append(dict_element)
+
+        else:
+            elements = [[], []]
+            for i, name in enumerate(list_name):
+                dict_element = {name_column_0: name}
+                for k, setting in enumerate(self.choice_variables):
+                    dict_element[setting] = list_value[k][i]
+
+                elements[0].append(dict_element)
+
+        # ScatterPlot
+        attributes = self.choice_variables
+        text_style = TextStyle(text_color=GREY,
+                               font_size=10,
+                               font_style='sans-serif')
+        surface_style = SurfaceStyle(color_fill=LIGHTGREY, opacity=0.3)
+        custom_tooltip = Tooltip(attributes=attributes,
+                                 surface_style=surface_style,
+                                 text_style=text_style,
+                                 tooltip_radius=10)
+
+        all_points = elements[0] + elements[1]
+
+        plots = []
+
+        # ScatterPlot
+        for j, setting in enumerate(list_settings):
+            for i in range(j + 1, len(list_settings)):
+                if len(plots) < 3:
+                    plots.append(Scatter(tooltip=custom_tooltip,
+                                         x_variable=list_settings[j],
+                                         y_variable=list_settings[i],
+                                         elements=all_points))
+
+        list_index_0 = [k for k in range(len(elements[0]))]
+        point_family_0 = PointFamily(LIGHTBLUE, list_index_0,
+                                     name='Non pareto')
+
+        n_pareto = len(elements[1])
+        list_index_1 = [len(all_points) - i - 1 for i in range(n_pareto)]
+        point_family_1 = PointFamily(LIGHTGREEN, list_index_1, name='Pareto')
+
+        # ParallelPlot
+        plots.append(ParallelPlot(elements=all_points,
+                                  axes=attributes))
+
+        return [MultiplePlots(plots=plots, elements=all_points, point_families=[point_family_0, point_family_1],
+                              initial_view_on=True)]
+
+
+def from_csv(filename: str, end: int = None, remove_duplicates: bool = False):
+    """
+    Generates MBSEs from given .csv file.
+    """
+    array = np.genfromtxt(filename, dtype=None, delimiter=',', names=True, encoding=None)
+    variables = [v for v in array.dtype.fields.keys()]
+    lines = []
+    for i, line in enumerate(array):
+        if end is not None and i >= end:
+            break
+        if not remove_duplicates or (remove_duplicates and line.tolist() not in lines):
+            lines.append(line.tolist())
+    return lines, variables
+
 
 def pareto_frontier(costs):
     """
     Find the pareto-efficient points
-
-    :param catalog: Catalog object on which to apply pareto_frontier
-                    computation
     :return: A (n_points, ) boolean array, indicating whether each point
              is Pareto efficient
     """
@@ -505,25 +513,7 @@ def pareto_frontier(costs):
     for index, cost in enumerate(costs):
         if is_efficient[index]:
             # Keep any point with a lower cost
-            is_efficient[is_efficient] = np.any(costs[is_efficient] < cost,
-                                                axis=1)
+            is_efficient[is_efficient] = np.any(costs[is_efficient] < cost, axis=1)
             # And keep self
             is_efficient[index] = True
     return is_efficient
-
-
-def from_csv(filename: str, end: int = None, remove_duplicates: bool = False):
-    """
-    Generates MBSEs from given .csv file.
-    """
-    array = np.genfromtxt(filename, dtype=None, delimiter=',',
-                          names=True, encoding=None)
-    variables = [v for v in array.dtype.fields.keys()]
-    lines = []
-    for i, line in enumerate(array):
-        if end is not None and i >= end:
-            break
-        if not remove_duplicates or (remove_duplicates
-                                     and line.tolist() not in lines):
-            lines.append(line.tolist())
-    return lines, variables
